# OBJ - timestamped JSON objects
#
# obj/base.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    providing timestamped JSON objects and their load/save methods.
    
"""

__txt__ = "Timestamped JSON Objects"
__version__ = 87
                 
## imports

import _thread
import collections
import datetime
import logging
import os
import importlib
import json
import sys
import time
import types
import threading

## defines

starttime = time.time()

## exceptions

class ECOMPAT(Exception):
    """ JSON on disk is not a dict. """
    pass

class EKEY(Exception):
    """ value is a function or method and is already present and cannot be overriden. """
    pass

class ENOTPATH(Exception):
    """ path is missing or not correct. """
    pass

## classes

class Object(object):

    """ The pure object that does load/save from and to JSON files. """

    def __repr__(self) -> str:
        """ module.class name repr. """
        return "%s.%s at %s" % (
            self.__class__.__module__,
            self.__class__.__name__,
            str(hex(id(self)))
        )

    def __str__(self) -> str:
        """ stringify on a 4 indent. """
        return json.dumps(self, default=smooth, indent=4, sort_keys=True)

    def cache(self, path):
        """ path, Object cache. """
        if path in cache:
            return cache[path]
        return self.load(path)

    def get(self, k, d=""):
        """ get attribute from object. """
        return getattr(self, k, d)

    def load(self, path: str) -> None:
        """ load an object from given path. """
        assert path
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        logging.debug("load %s" % path)
        with open(p) as f:
            data = json.load(f)
            if type(data) != dict:
                raise ECOMPAT(p)                
            for k, v in data.items():
                self.set(k, v)
        if "_container" not in dir(self):
            self._container = {}
        self._container["path"] = path
        cache[path] = self
        return self

    def search(self, m: {}) -> bool:
        """ return whether this object values matches criteria. """ 
        res = False
        for k in m.keys():
            v = self.get(k, None)
            if v != None:
                if m[k] in v: 
                    res = True
                else:
                    res = False
                    break
        return res

    def save(self, path="", stime="") -> str:
        """ save this object to a file. """
        if not path and not stime:
            try:
                path = self._container["path"]
            except (TypeError, AttributeError, KeyError):
                pass
        if not path:
            path = os.path.join(get_type(self), stime or rtime())
        if not path:
            raise ENOPATH
        if "_container" not in dir(self):
            self._container = {}
        self._container["path"] = path
        logging.info("save %s" % path)
        p = os.path.abspath(os.path.join(cfg.workdir, path)) 
        d, fn = os.path.split(p)
        if not os.path.isdir(d):
            cdir(d)
        with open(p, "w") as f:
            json.dump(self, f, default=smooth, indent=4, sort_keys=True)
        return path

    def set(self, k , v):
        """ set a attribute. """
        return setattr(self, k, v)

class Dotted(Object, dict):

    """ JSON load/saveable dict with dotted access (like a object). """

    def __getattribute__(self, name):
        try:
            return self[name]
        except KeyError as ex:
            return super().__getattribute__(name)

    def __getattr__(self, name):
        if name not in self:
            if "_default" in self:
                self[name] = self["_default"]
        try:
            return self[name]
        except KeyError:
            return super().__getattribute__(name)

    def __setattr__(self, key, value):
        if type(value) == types.MethodType:
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def last(self, m={}):
        store = Store()
        tmp = store.last(m, get_type(self))
        if tmp:
            self.update(tmp)

    def format(self, keys=["txt"], nr=None):
        """ return string with entries and optional txt. """
        oo = slice(self, keys)
        if nr != None:
            res = "%s " % nr
        else:
            res = ""
        for k in oo:
            res += self[k] + " "
        return res.rstrip()

    def upgrade(self, d):
        for k, v in d.items():
            if v:
                self[k] = v
        return self

    def upgrade2(self, d):
        for k, v in self.items():
            if k in d:
                if not v:
                    self[k] = d[k]

class Cache(Dotted):

    """ cache used to cache output to bots and other output objects. """

    def add(self, dest, txt):
        if "dest" not in self:
            self[dest] = []
        self[dest].append(txt)

    def more(self, dest, nr=10):
        if "dest" in dir(self):
            res = self[dest][-nr:]
            del self[dest][-nr:]    
            return res
        else:
            return []

class Cfg(Dotted):

    """ Base Config. """

    pass

class Event(Dotted):

    """ parse a string into a command callback combo. """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._default = ""
        self._ready = threading.Event()
        self._result = []
        if args:
            self.update(args[0])
        if kwargs:
            self.update(kwargs)

    def __getattr__(self, name):
        if name not in dir(self):
            self[name] = self._default
        return super().__getattr__(name)

    def get_cmd(self):
        v = self.get("txt", None)
        if v != None and v.strip():
            return v.strip().split()[0]
        return ""

    def ok(self, txt=""):
        self.reply("ok %s" % txt or self.get_cmd())

    def ready(self):
        self._ready.set()

    def reply(self, txt):
        self._result.append(txt)
        fleet.say(self.orig, self.channel, txt)
                    
    def parse(self):
        self._result = []
        self.args = []
        self.dkeys = []
        self.match = Dotted()
        self.skip = []
        self.types = []
        self.txt = self.txt.strip()
        self.cmd = self.get_cmd()
        splt = self.txt.split()
        nr = -1
        for arg in splt:
            nr += 1
            if nr == 0:
                self.cmd = arg
                continue
            if nr == 1:
                self.types.extend(match_types(arg))
            s = Selector(arg)
            if s:
                self.match.update(s)
                if arg.endswith("-"):
                    self.skip.append(arg.split("=")[0])
                if arg not in self.skip:
                    self.dkeys.append(arg.split("=")[0])
                self.args.append(arg)
                continue
            if nr >= 2:
                self.dkeys.append(arg)
            self.args.append(arg)
        self.rest = " ".join(self.args)
        if "-o" in self.rest:
            self.options = self.rest.split("-o")[-1].split()[0]
        if "-i" in self.rest:
            try:
                self.index = int(self.rest.split("-i")[-1].split()[0])
            except ValueError as ex:
                pass
        logging.debug(self)

    def wait(self):
        self._ready.wait()

class Fleet(Dotted):

    """ A bus can be used to send text to registered bots. """

    bots = []

    def __iter__(self):
        return self
        
    def __next__(self):
        for bot in self.bots:
            yield bot

    def add(self, bot):
        """ add a bots. """
        Fleet.bots.append(bot)

    def announce(self, txt):
        """ announce to all bots. """
        for bot in Fleet.bots:
             bot.announce(txt)
    
    def by_type(self, btype):
        """ return bots by type. """
        for bot in Fleet.bots:
            if str(btype).lower() in str(type(bot)).lower():
                return bot

    def get_bot(self, id):
        """ return a listener with corresponding id. """
        for bot in Fleet.bots:
            if id in repr(bot):
                return bot

    def match(self, m):
        """ search for matching bot. """
        for bot in Fleet.bots:
            if m.lower() in repr(bot):
                return bot

    def remove(self, bot):
        """ remove a listener. """
        Fleet.bots.remove(listener)

    def say(self, id, channel, txt):
        """ say some text to a listener. """
        bot = self.get_bot(id)
        if bot:
            bot.say(channel, txt)

class Register(Dotted):

    """ dict of list of values. """

    def register(self, k: str, v: object) -> None:
        """ register a key, value pair, constructing a list if key is not in object. """
        if k not in self:
            self[k] = []
        return self[k]

class Selector(Dotted):

    """ collects key=value pairs from a string. """

    def __init__(self, txt=""):
        super().__init__()
        if txt:
            self.parse(txt)

    def parse(self, txt):
        """ parse a string into a dict with key=value pairs. """
        txt = bytes(txt, "utf-8").decode("unicode_escape")
        quoted = False
        words = []
        key2 = ""
        for word in txt.split():
            try:
                key, value = word.split("=", 1)
            except (IndexError, ValueError):
                continue
            if value.startswith('"'):
                value = value[1:]
                key2 = key
                quoted = True
            if quoted and '"' in value:
                value = value.replace('"', '')
                quoted = False
                key = key2
            if quoted:
                words.append(value)
                continue
            else:
                words.append(value)
                value = " ".join(words)
            if key and value:
                if value.endswith("-"):
                    value = value[:-1]
                self[key] = value
        return self

class Store(Dotted):

    """ manages objects stored on disk. """

    def deleted(self, match={}):
        """ show objects with the deleted flag set. """
        for fn in filenames():
            o = construct(fn)
            if match and not o.search(match):
                continue
            if "_deleted" in dir(o) and o._deleted:
                yield o
        
    def find(self, match={}, type=None):
        """ find matching objects. """
        for fn in filenames(type):
            o = construct(fn)
            if not o:
                continue
            if "_deleted" in dir(o) and o._deleted:
                continue
            if type and not match:
                yield o
            elif o.search(match):
                yield o

    def last(self, match={}, type=None):
        """ show last object. """
        for fn in filenames(type)[::-1]:
            o = construct(fn)
            if not o:
                continue
            if "_deleted" in dir(o) and o._deleted:
                continue
            if type and not match:
                return o
            elif o.search(match):
                return o

## instances

cache = Cache()
cfg = Cfg()
cfg.workdir = os.path.join(os.path.expanduser("~"), ".obj")
fleet = Fleet()

## utility functions

def cdir(path) -> bool:
    """ create directory. """
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        p = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(p)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def construct(fn):
    """ construct a type and initialize it with data JSON dict. """
    try:
        cls = get_cls(get_fntype(fn))
    except ModuleNotFoundError:
        cls = Dotted
    try:
        return cls().cache(fn)
    except json.decoder.JSONDecodeError:
        logging.error("%s has decode errors." % fn)

def filenames(type=None):
    """ show file names of objects stored on path (index into the workdir). """
    res = []
    if type:
        wd = os.path.join(cfg.workdir, type)
    else:
        wd = cfg.workdir
    for rootdir, dirs, files in os.walk(wd, topdown=True):
        if wd not in rootdir:
            continue
        for fn in files:
            res.append(os.path.join(rootdir, fn).split(cfg.workdir)[-1][1:])
    return sorted(res, key=lambda x: fn_time(x))

def fn_time(daystr: str) -> None:
    """ return the time determined from the filename of an object. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.split(".")[0]
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    except ValueError:
        t = 0
    return t

def get_cls(name):
    """ return class from sys.modules. """
    name, c = name.rsplit(".", 1)
    mod = importlib.import_module(name)
    return getattr(mod, c)

def get_type(obj):
    """ return this type of object. """
    return obj.__class__.__module__ + "." + obj.__class__.__qualname__

def get_types():
    """ return all types available on disk. """
    return sorted(os.listdir(cfg.workdir))

def get_fntype(fn):
    """ return class from filename. """
    return fn.split(os.sep)[0]

def prefixes():
    """ show list of subdirectories in the workdir. """
    return sorted(set([x.split(".")[-1].lower() for x in os.listdir(cfg.workdir)]))

def match_types(name):
    names = []
    for t in get_types():
        if name.lower() in t.lower().split(".")[-1]:
            names.append(t)
    return names

def rtime():
    """ return filename base of current time. """ 
    return str(datetime.datetime.now()).replace(" ", os.sep)

def slice(o, keys):
    """ return subset of an object. """
    cls = get_cls(get_type(o))
    oo = cls()
    for k in keys:
        if k in o.keys():
           oo[k] = o[k]
    return oo 

def smooth(o):
    """ deal with objects that are not convertable to basic JSON types. """
    if isinstance(o, Object):
        return vars(o)
    if type(o) not in [dict, list, str, int, float, bool, None]:
        return repr(o)
    return str(o)
