# OBJ - timestamped JSON objects
#
# obj/shell.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" The OBJ Shell. """

## imports

import obj.base
import obj.handler
import obj.utils
import obj.users

import _thread
import atexit
import logging
import logging.handlers
import optparse
import os
import readline
import rlcompleter
import sys
import termios
import time

## defines

logdir = os.path.abspath(os.path.expanduser("~") + os.sep + ".objlogs")
histfile = os.path.join(logdir, "history")
resume = {}

class Defaults(obj.base.Dotted):

    """ Defaults for options. """

    pass

defaults = Defaults()
defaults.all = False
defaults.daemon = False
defaults.init = ""
defaults.level = ""
defaults.modules= ""
defaults.options = ""
defaults.owner = ""
defaults.shell = False
defaults.verbose = False
defaults.workdir = ""

opts = [
    ('-a', '', 'store_true', False, 'all', "load all modules."),
    ('-b', '', 'store_true', False, 'daemon', "switch to background mode."),
    ('-d', '', 'string', '', 'workdir', 'set working directory.'),
    ('-i', '', 'string', '', 'init', 'modules to initialize'),
    ('-l', '', 'string', '', 'level', 'loglevel.'),
    ('-m', '', 'string', '', 'modules', 'modules to load'),
    ('-o', '', "string", '', 'options', 'parseable options to use.'),
    ('-v', '', 'store_true', False, 'verbose', 'enable verbose mode.'),
    ('-z', '', 'store_true', False, 'shell', 'enable shell.'),
    ('', '--owner', "string", '', 'owner', 'owner of the bot'),
] 

## classes

class CLI(obj.handler.Handler):

    """ prompt less command line interface. """

    cache = obj.base.Cache()

    def __init__(self):
        super().__init__()
        self._sock = sys.stdout

    def announce(self, txt):
        """ announce to channel. """
        self.raw(txt)

    def cmd(self, txt):
        """ run command from text. """
        event = super().event(txt, "root@shell", options=obj.base.cfg.options)
        super().dispatch(event)
        return event

    def raw(self, txt):
        """ raw output on the socket. """
        self._sock.write(str(txt))
        self._sock.write("\n")
        self._sock.flush()

    def say(self, channel, txt):
        """ convert channel output to txt. """
        self.cache.add(channel, txt)
        self.raw(txt)

    def wait_for_queue(self):
        """ wait for the queue to be empty. """
        while 1:
            time.sleep(0.01)
            s = self.queue.qsize()
            if s == 0:
                break

class Shell(CLI):

    """ Provides interface to a command line interface. """    

    def dispatch(self, event):
        """ display a prompt after dispatch. """
        super().dispatch(event)
        self.prompt()

    def get_event(self):
        """ return a string from keyboard input. """
        return self.event(input())

    def prompt(self):
        """ show prompt. """
        self._sock.write("> ")
        self._sock.flush()

class TestBot(CLI):

    def __init__(self):
        super().__init__()
        self.verbose = False

    def raw(self, txt):
        """ raw output on the console. """
        if not self.verbose:
            return
        super().raw(txt)

## log related classes

class Completer(rlcompleter.Completer):

    """ enables tab completion for the shell. """

    def __init__(self, options=[]):
        super().__init__()
        self.commands = options or sorted(obj.handler.names.keys())

    def complete(self, text, state):
        """ suggest tab completion. """
        if state == 0:
            if text:
                self.matches = [s for s in self.commands if s and s.startswith(text)]
            else:
                self.matches = self.commands[:]
        if state < len(self.commands):
            return self.matches[state]
        else:
            return None

    def display(self, substitution, matches, depth):
        """ display options to complete. """
        if matches:
            print("\t".join(matches))
            readline.redisplay()
    
class DumpHandler(logging.StreamHandler):

    """ A logging handler that ignores the log messages. """

    def emit(self, record):
        """ neglect logging record emission. """
        pass

## functions

def close_history() -> None:
    """ close the history file (path is in histfile variable). """
    readline.write_history_file(histfile)

def daemon() -> None:
    """ switch to background mode. """
    pid = os.fork()
    if pid != 0:
        os._exit(0)
    os.chdir("/")
    os.setsid()
    os.umask(0)
    pid = os.fork()
    if pid != 0:
        os._exit(0)
    path = os.path.abspath(os.path.join(logdir, "pidfile"))
    f = open(path, 'w')
    f.write(str(os.getpid()))
    f.flush()
    f.close()
    sys.argv[0] = "obj"
    sys.stdout.flush()
    sys.stderr.flush()
    si = open("/dev/null", 'r')
    so = open("/dev/null", 'a+')
    se = open("/dev/null", 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

def enable_history() -> None:
    """" enable logging of history to the history file (path is in histfile variable). """
    if not os.path.exists(histfile):
        d, f = os.path.split(histfile)
        obj.utils.cdir(d)
        touch(histfile)
    readline.read_history_file(histfile)
    atexit.register(close_history)

def get_completer() -> Completer:
    """ return a completer. """
    return readline.get_completer()

def level(level, nostream=False, always=True):
    """ set loglevel. """
    datefmt = '%H:%M:%S'
    format_time = "%(asctime)-8s %(message)-70s"
    format_plain = "%(message)-0s"
    level = level.upper()
    logger = logging.getLogger("")
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    try:
        logger.setLevel(level)
    except ValueError:
        return
    formatter = logging.Formatter(format_plain, datefmt)
    if nostream:
        dhandler = DumpHandler()
        dhandler.propagate = False
        dhandler.setLevel(level)
        logger.addHandler(dhandler)
    else:
        ch = logging.StreamHandler()
        ch.propagate = False
        ch.setFormatter(formatter)
        ch.setLevel(level)
        logger.addHandler(ch)
    formatter2 = logging.Formatter(format_time, datefmt)
    filehandler = logging.handlers.TimedRotatingFileHandler(os.path.join(logdir, "log"), 'midnight')
    filehandler.propagate = False
    filehandler.setFormatter(formatter2)
    if always:
        filehandler.setLevel(logging.DEBUG)
    else:
        filehandler.setLevel(level)
    logger.addHandler(filehandler)
    return logger

def main(name="obj", version=obj.base.__version__, options=[], defs={}, wait=True):
    """ main program. """
    opts.extend(options)
    cfg = parse_cli(name, version, sorted(opts), defs)
    if not cfg.workdir:
        cfg.workdir = os.path.abspath(os.path.join(os.path.expanduser("~"), ".obj"))
    obj.base.cfg.update(cfg)
    if not os.path.exists(cfg.workdir):
        obj.utils.cdir(cfg.workdir)
    if cfg.daemon:
        path = os.path.join(logdir, "pidfile")
        f = open(path, 'w')
        f.write(str(os.getpid()))
        f.flush()
        f.close()
        daemon()
        level(cfg.level or "debug")
        logging.warn("%s #%s started at %s" % (name.upper(), obj.base.__version__, time.ctime(time.time())))
        bot = CLI()
    else:
        level(cfg.level or "error")
        bot = Shell()
    if cfg.owner:
        users = obj.users.Users()
        users.oper(cfg.owner)
    for n in [x for x in cfg.modules.split(",") if x]:
        bot.load_mod(n)
    for n in [x for x in cfg.init.split(",") if x]:
        nn = "obj.server.%s" % n
        try:
            m = bot.load_mod(nn)
        except ModuleNotFoundError:
            continue
        try:
            m.start()
        except Exception as ex:
            logging.error(obj.utils.get_exception())
    set_completer()
    bot.walk("obj.cmds")
    bot.start()
    if cfg.args and not cfg.shell:
        bot.cmd(" ".join(cfg.args))
        return bot, cfg
    else:
        if not cfg.daemon and not cfg.args or cfg.shell:
            bot.prompt()
    if wait:
        try:
            bot.wait()
        except:
            bot.stop()
    return bot, cfg

def make_opts(options, usage, version) -> list:
    """ create commandline options. """
    parser = optparse.OptionParser(usage=usage, version=version)
    for option in options:
        optiontype, default, dest, helptype = option[2:]
        if "store" in optiontype:
            try:
                parser.add_option(option[0], option[1], action=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("%s option %s" % (str(ex), option))
                continue
        else:
            try:
                parser.add_option(option[0], option[1], type=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("^%s option %s" % (str(ex), option))
                continue
    args = parser.parse_args()
    return args

def parse_cli(name, version=obj.base.__version__, options=[], defs={}, usage="", workdir="") -> obj.base.Dotted:
    """ parse command line options. """
    usage = usage or "%s [options]" % name
    ver = "%s #%s" % (name.upper(), version)
    opt, arguments = make_opts(options or opts, usage, ver)
    cfg = obj.base.Dotted(vars(opt))
    if defs:
        defaults.update(defs)
    if cfg != defaults:
        cfg.changed = True
    else:
        cfg.changed = False
    cfg.args = arguments
    if workdir:
        cfg.workdir = workdir
    return cfg

def reset() -> None:
    """ reset terminal to previous state (prevents garblement on console). """
    close_history()
    if "old" in resume:
        try:
            termreset(resume["fd"], resume["old"])
        except:
            pass

def set_completer(args=[]) -> None:
    """ set a list of options in the completer. """
    completer = Completer(args)
    readline.set_completer_delims(' ')
    readline.parse_and_bind("tab: complete")
    readline.set_completion_display_matches_hook(completer.display)
    readline.set_completer(completer.complete)
    atexit.register(lambda: readline.set_completer(None))

def signal_term(sig, frame) -> None:
    """ handle the sighup signal, calls reload on all bus listeners. """
    for listener in bus.listeners:
         listener.reload()

def startup() -> None:
    """ run this function on startup, save the console state. """
    global resume
    enable_history()
    resume["fd"] = sys.stdin.fileno()
    resume["old"] = termsetup(sys.stdin.fileno())
    atexit.register(reset)

def termreset(fd, old) -> None:
    """ do an actual terminal reset, restores old state. """ 
    termios.tcsetattr(fd, termios.TCSADRAIN, old)

def termsetup(fd) -> list:
    """ return terminal setup. """
    old = termios.tcgetattr(fd)
    return old

def touch(fname) -> None:
    """ touch a file, creates a file without any content. """
    try:
        fd = os.open(fname, os.O_RDONLY | os.O_CREAT)
        os.close(fd)
    except TypeError:
        pass
