# OBJ - timestamped JSON objects
#
# obj/utils.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 18-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    utility module.
    
"""

__version__ = 1

## imports

import _thread
import datetime
import importlib
import logging
import obj.base
import os
import stat
import sys
import time
import traceback

## defines

timestrings = [
               "%a, %d %b %Y %H:%M:%S %z",
               "%d %b %Y %H:%M:%S %z",
               "%d %b %Y %H:%M:%S",
               "%a, %d %b %Y %H:%M:%S",
               "%d %b %a %H:%M:%S %Y %Z",
               "%d %b %a %H:%M:%S %Y %z",
               "%a %d %b %H:%M:%S %Y %z",
               "%a %b %d %H:%M:%S %Y",
               "%d %b %Y %H:%M:%S",
               "%a %b %d %H:%M:%S %Y",
               "%Y-%m-%d %H:%M:%S",
               "%Y-%m-%dt%H:%M:%S+00:00",
               "%a, %d %b %Y %H:%M:%S +0000",
               "%d %b %Y %H:%M:%S +0000",
              ]

# 2012-10-23 00:00:00
year_formats = [
    "%b %H:%M",
    "%b %H:%M:%S",
    "%a %H:%M %Y",
    "%a %H:%M",
    "%a %H:%M:%S",
    "%Y-%m-%d",
    "%d-%m-%Y",
    "%d-%m",
    "%m-%d",
    "%Y-%m-%d %H:%M:%S",
    "%d-%m-%Y %H:%M:%S",
    "%d-%m %H:%M:%S",
    "%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M",
    "%d-%m-%Y %H:%M",
    "%d-%m %H:%M",
    "%m-%d %H:%M",
    "%H:%M:%S",
    "%H:%M"
]

## exceptions

class ENODATE(Exception):

    """ a date could not be determined. """

    pass

## functions

def cdir(path) -> bool:
    """ create direcotry. """
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        p = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(p)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def check_permissions(p, dirmask=0o700, filemask=0o600) -> None:
    """ check if directory permissions are set to readonly and if not set them to readonly. """
    uid = os.getuid()
    gid = os.getgid()
    try:
        stats = os.stat(p)
    except FileNotFoundError:
        return
    except OSError:
        d, fn = os.path.split(p)
        cdir(d)
        stats = os.stat(d)
    if stats.st_uid != uid:
        os.chown(p, uid, gid)
    if os.path.isfile(p):
        mask = filemask
    else:
        mask = dirmask
    m = oct(stat.S_IMODE(stats.st_mode))
    if m != oct(mask):
        os.chmod(p, mask)

def construct(type, data={}, origin=""):
    """ construct a type and initialize it with data JSON dict. """
    if origin:
        data["origin"] = origin
    try:
        cls = get_cls(type)
    except ModuleNotFoundError:
        cls = obj.base.Dotted
    if cls:
        try:
            return cls(data)
        except TypeError:
            return cls()

def days(o):
    """ show days till creation. """
    return elapsed(time.time() - fn_time(o._container.get("path", 0)))

def elapsed(seconds, short=True):
    """ show elapsed time. """
    txt = ""
    sub = str(seconds).split(".")[-1]
    nsec = float(seconds)
    year = 365*24*60*60
    week = 7*24*60*60
    day = 24*60*60
    hour = 60*60
    minute = 60
    years = int(nsec/year)
    nsec -= years*year
    weeks = int(nsec/week)
    nsec -= weeks*week
    days = int(nsec/day)
    nsec -= days*day
    hours = int(nsec/hour)
    nsec -= hours*hour
    minutes = int(nsec/minute)
    sec = nsec - minutes*minute
    if years:
        txt += "%sy" % years
    if weeks:
        days += weeks * 7
    if days:
        txt += "%sd" % days
    if years and short and txt:
        return txt
    if hours:
        txt += "%sh" % hours
    if days and short and txt:
        return txt
    if minutes:
        txt += "%sm" % minutes
    if hours and short and txt:
        return txt
    if sec == 0:
        txt += "0s"
    elif sec < 1 or not short:
        txt += "%.3fs" % sec
    else:
        txt += "%ss" % int(sec)
    txt = txt.strip()
    return txt

def fn_time(daystr: str) -> None:
    """ return the time determined from the filename of an object. """
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    datestr = datestr.split(".")[0]
    try:
        t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    except ValueError:
        t = 0
    return t

def get_cls(name):
    try:
        modname, clsname = name.rsplit(".", 1)
    except:
        return None
    try:
        mod = importlib.import_module(modname)
        o = getattr(mod, clsname)
    except:
        logging.error(get_exception())
        o = obj.base.Dotted()
    return o

def get_exception(txt="") -> str:
    """ return exception from stack. """
    exctype, excvalue, tb = sys.exc_info()
    trace = traceback.extract_tb(tb)
    result = ""
    for i in trace:
        fname = i[0]
        linenr = i[1]
        func = i[2]
        plugfile = fname[:-3].split(os.sep)
        mod = []
        for i in plugfile[::-1]:
            mod.append(i)
            if i == "obj":
                break
        ownname = '.'.join(mod[::-1])
        result += "%s:%s %s | " % (ownname, linenr, func)
    del trace
    res = "%s%s: %s %s" % (result, exctype, excvalue, str(txt))
    return res

def get_time(daystr):
    for f in year_formats:
        try:
            t = time.mktime(time.strptime(daystr, f))
            return t
        except:
            pass

def locked(func, *args, **kwargs):

    """ lock a method or function. """
    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        """ encapsulate the function into a blocking one. """
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except:
                pass
        return res
    return lockedfunc

def name(obj):
    """ return a classname.methodname of a function. """
    try:
        n = "%s.%s" % (obj.__self__.__class__.__name__, obj.__name__)
    except:
        try:
            n = "%s.%s" % (obj.__class__.__name__, obj.__name__)
        except:
            try:
                n = obj.__class__.__name__
            except:
                n = obj.__name__
    return n

def rtime():
    res =  str(datetime.datetime.now()).replace(" ", os.sep)
    return res

def scan_time(daystring):
    d = time.localtime()
    tstr = ""
    dstr = ""
    ystr = ""
    wday = ""
    mstr = ""
    for word in daystring.split():
        if word in weekdays:
            wday = word
        elif word in months:
            month = word
        elif "-" in word:
            dstr = word
        elif ":" in word:
            tstr = word
    if not tstr:
        tstr = str(d.tm_hour) + ":" + str(d.tm_min)
    if wday:
        difftonow = weekdays.index(wday) - d.tm_wday
        if difftonow < 0:
            difftonow = 7 + difftonow
        d = time.localtime(time.mktime(time.localtime()) + difftonow * 24* 60 * 60)
    if not dstr:
        dstr = str(d.tm_mday) + "-" + str(d.tm_mon) + "-" + str(d.tm_year)
    line = dstr + " " + tstr
    try:
        return get_time(line.strip())
    except ValueError:
        pass

def stripped(jid):
    """ remove everything after a '/' """
    try:
        return str(jid).split("/")[0]
    except:
        return str(jid)

def today():
    """ today at 0:00. """
    t = rtime().split(".")[0]
    ttime = time.strptime(t, "%Y-%m-%d/%H:%M:%S")
    result = time.mktime(ttime)
    return result

def to_day(daystring):
    previous = ""
    line = ""
    daystr = str(daystring)
    for word in daystr.split():
        if "-" in word:
            line += word + " "
        elif ":" in word:
            line += word
    if "-" not in line:
        line = day() + " " + line
    try:
        return get_time(line.strip())
    except ValueError:
        pass

def to_time(daystr):
    """ convert a string to a unix timestamp. """
    daystr = daystr.strip()
    if "," in daystr:
        daystr = " ".join(daystr.split(None)[1:7])
    elif "(" in daystr:
        daystr = " ".join(daystr.split(None)[:-1])
    else:
        try:
            d, h = daystr.split("T")
            h = h[:7]
            daystr = " ".join([d, h])
        except:
            pass
    res = 0
    for tstring in timestrings:
        try:
            res = time.mktime(time.strptime(daystr, tstring))
        except:
            try:
                res = time.mktime(time.strptime(" ".join(daystr.split()[:-1]), tstring))
            except:
                pass
        if res:
           break
    if not res:
        raise ENODATE(daystr)
    return res
